package org.briarproject.bramble.mailbox;

import org.briarproject.bramble.api.contact.ContactManager;
import org.briarproject.bramble.api.contact.PrivateMailbox;
import org.briarproject.bramble.api.contact.event.ContactAddedEvent;
import org.briarproject.bramble.api.db.DatabaseComponent;
import org.briarproject.bramble.api.db.DatabaseExecutor;
import org.briarproject.bramble.api.db.DbException;
import org.briarproject.bramble.api.db.Transaction;
import org.briarproject.bramble.api.event.Event;
import org.briarproject.bramble.api.event.EventBus;
import org.briarproject.bramble.api.event.EventListener;
import org.briarproject.bramble.api.lifecycle.IoExecutor;
import org.briarproject.bramble.api.mailbox.MailboxInfo;
import org.briarproject.bramble.api.mailbox.MailboxManager;
import org.briarproject.bramble.api.mailbox.MailboxService;
import org.briarproject.bramble.api.plugin.ConnectionRegistry;
import org.briarproject.bramble.api.plugin.LanTcpConstants;
import org.briarproject.bramble.api.plugin.PluginManager;
import org.briarproject.bramble.api.plugin.TorConstants;
import org.briarproject.bramble.api.plugin.duplex.DuplexPlugin;
import org.briarproject.bramble.api.plugin.duplex.DuplexTransportConnection;
import org.briarproject.bramble.api.properties.TransportProperties;
import org.briarproject.bramble.api.properties.TransportPropertyManager;
import org.briarproject.bramble.api.system.Scheduler;

import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.Executor;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Logger;

import javax.inject.Inject;

import static java.util.concurrent.TimeUnit.MILLISECONDS;
import static java.util.logging.Level.INFO;
import static java.util.logging.Level.WARNING;
import static org.briarproject.bramble.api.mailbox.MailboxConstants.MIN_POLLING_INTERVALL;
import static org.briarproject.bramble.util.LogUtils.logException;

/**
 * The Mailbox Service runs to poll mailboxes
 */
public class MailboxServiceImpl implements MailboxService, EventListener {
	private static final Logger LOG =
			Logger.getLogger(MailboxServiceImpl.class.getName());

	private final Executor ioExecutor;
	private Executor dbExecutor;
	private final ScheduledExecutorService scheduler;
	private EventBus eventBus;
	private ContactManager contactManager;
	private final DatabaseComponent db;
	private final ConnectionRegistry connectionRegistry;
	private final PluginManager pluginManager;
	private final MailboxManager mailboxManager;
	private final TransportPropertyManager transportPropertyManager;
	private SecureRandom secureRandom;


	private volatile Future mailboxFuture;
	private AtomicBoolean hasPrivateMailbox = new AtomicBoolean(false);
	private PrivateMailbox privateMailbox;

	@Inject
	MailboxServiceImpl(@IoExecutor Executor ioExecutor,
			@DatabaseExecutor Executor dbExecutor, @Scheduler
			ScheduledExecutorService scheduler, EventBus eventBus,
			ContactManager contactManager,
			DatabaseComponent db,
			ConnectionRegistry connectionRegistry,
			PluginManager pluginManager, MailboxManager mailboxManager,
			TransportPropertyManager transportPropertyManager,
			SecureRandom secureRandom) {
		this.ioExecutor = ioExecutor;
		this.dbExecutor = dbExecutor;
		this.scheduler = scheduler;
		this.eventBus = eventBus;
		this.contactManager = contactManager;
		this.db = db;
		this.connectionRegistry = connectionRegistry;
		this.pluginManager = pluginManager;
		this.mailboxManager = mailboxManager;
		this.transportPropertyManager = transportPropertyManager;
		this.secureRandom = secureRandom;
	}

	@Override
	public void startService() {
		if (LOG.isLoggable(INFO))
			LOG.info("Starting Mailbox Service");

		// Get private mailbox contact if available
		dbExecutor.execute(() -> getPrivateMailbox());

		mailboxFuture =
				schedule(new ConnectMailboxesTask(MIN_POLLING_INTERVALL),
						MIN_POLLING_INTERVALL);

		this.eventBus.addListener(this);
	}

	private Future schedule(Runnable task, int delay) {
		return scheduler
				.schedule(() -> ioExecutor.execute(task), delay, MILLISECONDS);
	}

	@Override
	public void eventOccurred(Event e) {
		if (e instanceof ContactAddedEvent) {
			if (hasPrivateMailbox.get())
				return;
			dbExecutor.execute(() -> getPrivateMailbox());
		}
	}

	private void getPrivateMailbox() {
		try {
			privateMailbox = contactManager.getPrivateMailbox();
		} catch (DbException e) {
			if (LOG.isLoggable(WARNING))
				LOG.info(e.toString());
		}
		if (privateMailbox != null)
			hasPrivateMailbox.getAndSet(true);
	}

	private class ConnectMailboxesTask implements Runnable {
		private int delay;

		private ConnectMailboxesTask(int min_interval) {
			this.delay = (int) (min_interval +
					secureRandom.nextFloat() * min_interval);
		}

		void reschedule() {
			mailboxFuture = schedule(this, delay);
		}

		@Override
		public void run() {
			// NOTE: Using Lan then Tor order (this is debatable but
			// surely useful for testing)
			ArrayList<DuplexPlugin> activePlugins = new ArrayList<>();

			DuplexPlugin p =
					(DuplexPlugin) pluginManager.getPlugin(LanTcpConstants.ID);
			if (p != null && p.isRunning())
				activePlugins.add(p);

			p =	(DuplexPlugin) pluginManager.getPlugin(TorConstants.ID);
			if (p != null && p.isRunning())
				activePlugins.add(p);

			// Wait for plugins to become available
			if (activePlugins.isEmpty())
				reschedule();

			// Try to connect to private mailbox
			try {
				connectPrivateMailbox(activePlugins);
			} catch (DbException e) {
				logException(LOG, WARNING, e);
				return;
			}

			// Poll Contact mailboxes
			Collection<MailboxInfo> contactMailboxes;
			Transaction txn = null;
			try {
				txn = db.startTransaction(true);
				contactMailboxes = db.getContactMailboxes(txn);
				db.commitTransaction(txn);
			} catch (DbException e) {
				logException(LOG, WARNING, e);
				return;
			} finally {
				if (txn != null)
					db.endTransaction(txn);
			}

			for (MailboxInfo mailboxInfo : contactMailboxes) {
				// Don't connect to mailbox if contact is currently connected
				if (connectionRegistry
						.isConnected(mailboxInfo.getContactId()))
					continue;

				// If mailboxId == null the contact does not have their own mailbox
				// but has been introduced to our mailbox
				if (mailboxInfo.getMailboxId() == null) {
					mailboxManager
							.handleContactWithoutMailbox(mailboxInfo);
					continue;
				}

				ioExecutor.execute(() -> connectContactMailbox(activePlugins,
						mailboxInfo));
			}

			reschedule();
		}

		private void connectContactMailbox(
				ArrayList<DuplexPlugin> activePlugins,
				MailboxInfo mailboxInfo) {

			for (DuplexPlugin plugin : activePlugins) {

				TransportProperties properties;
				try {
					properties = transportPropertyManager
							.getRemoteProperties(mailboxInfo.getMailboxId(),
									plugin.getId());
				} catch (DbException e) {
					logException(LOG, WARNING, e);
					return;
				}

				DuplexTransportConnection conn =
						plugin.createConnection(properties);

				if (conn != null) {
					if (LOG.isLoggable(INFO))
						LOG.info("Contact mailbox connected");

					mailboxManager.handleOutgoingContactMailboxConnection(
							mailboxInfo.getContactId(),
							mailboxInfo.getMailboxId(),
							conn, plugin.getId());
					return;
				}
			}
		}

		private void connectPrivateMailbox(
				ArrayList<DuplexPlugin> activePlugins) throws DbException {
			// Check if mailbox is available and if it is connected
			if (!hasPrivateMailbox.get() ||
					connectionRegistry.isConnected(privateMailbox.getId()))
				return;

			for (DuplexPlugin plugin : activePlugins) {

				TransportProperties properties = transportPropertyManager
						.getRemoteProperties(privateMailbox.getId(),
								plugin.getId());

				if (LOG.isLoggable(INFO))
					LOG.info("Connecting to private mailbox");

				DuplexTransportConnection conn =
						plugin.createConnection(properties);


				if (conn != null) {
					if (LOG.isLoggable(INFO))
						LOG.info("Connected to private mailbox");
					mailboxManager
							.handleOutgoingPrivateMailboxConnection(
									privateMailbox.getId(), conn,
									plugin.getId());
					return;
				}
			}

			if (LOG.isLoggable(INFO))
				LOG.info("Unable to connect to private mailbox");
		}
	}

	@Override
	public void stopService() {
		mailboxFuture.cancel(true);
		eventBus.removeListener(this);
	}
}
