package org.briarproject.bramble.mailbox.protocol;

import org.briarproject.bramble.api.data.BdfList;
import org.briarproject.bramble.api.data.BdfReader;
import org.briarproject.bramble.api.data.BdfWriter;
import org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE;

import java.io.EOFException;
import java.io.IOException;
import java.net.ProtocolException;
import java.util.LinkedHashMap;
import java.util.Map.Entry;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Executor;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Logger;

import static java.util.logging.Level.INFO;
import static java.util.logging.Level.WARNING;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.RESPONSE;

public class MailboxProtocol implements Runnable {


	private static final Logger LOG =
			Logger.getLogger(MailboxProtocol.class.getName());

	private final Executor ioExecutor;
	private final BdfWriter mailboxBdfWriter;
	private final BdfReader mailboxBdfReader;

	volatile BlockingQueue<MailboxMessage> outQueue =
			new LinkedBlockingQueue<>();

	volatile BlockingQueue<MailboxRequest> inQueue =
			new LinkedBlockingQueue<>();

	volatile ConcurrentHashMap<Long, MailboxRequest> pendingRequests =
			new ConcurrentHashMap<>();

	private LinkedHashMap<TYPE, MailboxRequestHandler> requestHandlers =
			new LinkedHashMap<>();

	private volatile AtomicBoolean stopped = new AtomicBoolean(false);
	private volatile Thread writingThread = null;


	public MailboxProtocol(Executor ioExecutor, BdfWriter mailboxBdfWriter,
			BdfReader mailboxBdfReader) {
		this.ioExecutor = ioExecutor;
		this.mailboxBdfWriter = mailboxBdfWriter;
		this.mailboxBdfReader = mailboxBdfReader;

	}

	public void sendKeepAlive() throws IOException {
		// flush the writer without pending data to send a keepalive
		if (stopped.get())
			throw new ProtocolException("Mailbox protocol has stopped");
		synchronized (mailboxBdfWriter) {
			mailboxBdfWriter.flush();
		}
	}

	public void writeRequest(MailboxRequest req)
			throws InterruptedException, IOException {
		if (req.hasResponse())
			pendingRequests.put(req.getId(), req);
		writeMailboxMessage(req);

	}

	public void writeSucessResponse(MailboxRequest req)
			throws InterruptedException, IOException {
		if (!req.hasResponse())
			throw new RuntimeException(
					"Trying to send a response for a request that does not expect a response");
		writeMailboxMessage(req.createSuccessResponse());
	}

	public void writeErrorResponse(MailboxRequest req, String error)
			throws InterruptedException, IOException {
		if (!req.hasResponse())
			throw new RuntimeException(
					"Trying to send a response for a request that does not expect a response");

		writeMailboxMessage(req.createErrorResponse(error));
	}

	private void writeMailboxMessage(MailboxMessage msg)
			throws InterruptedException, ProtocolException {
		synchronized (outQueue) {
			if (stopped.get())
				throw new ProtocolException("Protocol has stopped");
			outQueue.put(msg);
		}
	}

	/**
	 * Run the mailbox protocol. This function terminates when the underlying
	 * input/output streams are closed
	 */
	@Override
	public void run() {
		ioExecutor.execute(() -> readIncomingMessages());
		writeOutgoingMessages();
	}

	/**
	 * Register a Handler for a specific request type.
	 * <p>
	 * NOTE: {@link this#run()} MUST be called AFTER all handlers
	 * have been registered!
	 *
	 * @param handler
	 */
	public void registerRequestHandler(MailboxRequestHandler handler) {
		if (writingThread != null)
			throw new RuntimeException(
					"Trying to register request handler after run call!");
		if (requestHandlers.containsKey(handler.getType()))
			throw new RuntimeException(
					"Handler for " + handler.getType().toString() +
							" has been registered twice!");
		requestHandlers.put(handler.getType(), handler);
	}

	private void readIncomingMessages() {
		BdfList bdfMsg;

		while (!stopped.get()) {
			MailboxMessage mailboxMessage;
			try {
				if (mailboxBdfReader.eof())
					throw new EOFException();
				bdfMsg = mailboxBdfReader.readList();
			} catch (IOException e) {
				handleIOException(true, e);
				return;
			}

			try {
				mailboxMessage = MailboxMessage.parse(bdfMsg);
			} catch (ProtocolException e) {
				LOG.info(e.toString());
				continue;
			}

			if (mailboxMessage.getType() == RESPONSE)
				handleResponse((MailboxResponse) mailboxMessage);
			else
				ioExecutor.execute(() -> handleRequest(
						(MailboxRequest) mailboxMessage));
		}
	}

	private void handleRequest(MailboxRequest req) {
		// TODO: Limit number of requests which are handled parallel
		String error = null;
		MailboxRequestHandler handler =
				requestHandlers.get(req.getType());
		if (handler != null) {
			try {
				handler.handleRequest(req);
			} catch (ProtocolException e) {
				if (!req.hasResponse())
					throw new RuntimeException(
							"Protocol exception was thrown for request without response");

				error = e.toString();
			}
		} else {
			error = "Unsupported Request";
			if (LOG.isLoggable(WARNING))
				LOG.warning("Received unsupported Request: " +
						req.getType().name());
		}

		if (!req.hasResponse())
			return;

		try {
			if (error == null)
				writeSucessResponse(req);
			else
				writeErrorResponse(req, error);
		} catch (InterruptedException | IOException e) {
			if (LOG.isLoggable(INFO))
				LOG.info(e.toString());
		}
	}

	private void handleResponse(MailboxResponse msg) {
		MailboxRequest req = pendingRequests.remove(msg.getId());

		if (req == null) {
			if (LOG.isLoggable(WARNING))
				LOG.warning("Received response not matching with any request");
			return;
		}

		if (msg.isSuccess())
			req.signalSucess();
		else
			req.signalError(msg.getErrorMessage());

	}

	private void writeOutgoingMessages() {
		writingThread = Thread.currentThread();
		MailboxMessage message;

		while (!stopped.get()) {
			try {
				message = outQueue.take();
			} catch (InterruptedException e) {
				writingThread.interrupt();
				handleIOException(false,
						new IOException("Writer Interrupted"));
				return;
			}

			synchronized (mailboxBdfWriter) {
				try {
					mailboxBdfWriter.writeList(message.toBdfList());
					mailboxBdfWriter.flush();
				} catch (IOException e) {
					writingThread.interrupt();
					handleIOException(false, e);
					return;
				}
			}
		}
	}

	private void handleIOException(boolean isReadingThread, IOException e) {
		if (isReadingThread && writingThread != null)
			writingThread.interrupt();

		// Exception occurred after the protocol was stopped externally
		if (stopped.getAndSet(true))
			return;

		if (LOG.isLoggable(INFO))
			LOG.info("Mailbox Protocol was closed: " + e.toString());

		// Signal that the session has ended to a thread waiting for requests
		inQueue.clear();
		inQueue.add(new MailboxRequestEnd());

		// Ensure that threads waiting to write a request terminate
		outQueue.clear();

		// Signal the error to anyone waiting for a response
		for (Entry<Long, MailboxRequest> entry : pendingRequests.entrySet()) {
			MailboxResponse r = new MailboxResponse(entry.getKey(),
					"Connection closed");
			entry.getValue().signalError(r.getErrorMessage());
		}

		// notify all handlers
		for (MailboxRequestHandler handler : requestHandlers.values())
			handler.protocolFinished();
	}
}
