package org.briarproject.bramble.mailbox.sessions;

import org.briarproject.bramble.api.contact.ContactId;
import org.briarproject.bramble.api.db.DatabaseComponent;
import org.briarproject.bramble.api.db.DbException;
import org.briarproject.bramble.api.db.Transaction;
import org.briarproject.bramble.api.mailbox.MailboxConstants;
import org.briarproject.bramble.api.mailbox.MailboxSession;
import org.briarproject.bramble.api.sync.SyncSession;
import org.briarproject.bramble.api.sync.SyncSessionFactory;
import org.briarproject.bramble.api.transport.KeyManager;
import org.briarproject.bramble.api.transport.StreamContext;
import org.briarproject.bramble.api.transport.StreamReaderFactory;
import org.briarproject.bramble.api.transport.StreamWriter;
import org.briarproject.bramble.api.transport.StreamWriterFactory;
import org.briarproject.bramble.mailbox.protocol.MailboxMessage;
import org.briarproject.bramble.mailbox.protocol.MailboxProtocol;
import org.briarproject.bramble.mailbox.protocol.MailboxRequest;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestEnd;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestHandler;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestSync;

import java.io.ByteArrayOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.Executor;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Logger;

import javax.annotation.Nullable;

import static java.util.logging.Level.INFO;
import static java.util.logging.Level.WARNING;
import static org.briarproject.bramble.api.transport.TransportConstants.TAG_LENGTH;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.END;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.SYNC;

public abstract class AbstractMailboxSession implements MailboxSession {
	private static final Logger LOG =
			Logger.getLogger(AbstractMailboxSession.class.getName());
	private final Executor ioExecutor;
	private final ContactId contactId;
	private DatabaseComponent db;
	private KeyManager keyManager;
	private SyncSessionFactory syncSessionFactory;
	private final StreamWriterFactory streamWriterFactory;
	private final StreamReaderFactory streamReaderFactory;
	private MailboxProtocol mailboxProtocol;
	private final int transportMaxLatency;
	private final int transportMaxIdleTime;

	// Used to handle graceful session termination with END request
	private AtomicBoolean remoteSessionFinished = new AtomicBoolean(false);

	@Nullable
	private SYNCHandler syncHandler = null;


	public AbstractMailboxSession(Executor ioExecutor,
			DatabaseComponent db,
			KeyManager keyManager,
			SyncSessionFactory syncSessionFactory,
			StreamWriterFactory streamWriterFactory,
			StreamReaderFactory streamReaderFactory,
			MailboxProtocol mailboxProtocol,
			int transportMaxLatency,
			int transportMaxIdleTime,
			ContactId contactId) {

		this.ioExecutor = ioExecutor;
		this.db = db;
		this.keyManager = keyManager;
		this.syncSessionFactory = syncSessionFactory;
		this.streamWriterFactory = streamWriterFactory;
		this.streamReaderFactory = streamReaderFactory;
		this.mailboxProtocol = mailboxProtocol;
		this.transportMaxLatency = transportMaxLatency;
		this.transportMaxIdleTime = transportMaxIdleTime;
		this.contactId = contactId;

		// All sessions must enable END handling
		mailboxProtocol.registerRequestHandler(new ENDHandler());
	}

	public void endSession()
			throws IOException {

		try {
			mailboxProtocol.writeRequest(new MailboxRequestEnd());
			synchronized (remoteSessionFinished) {
				while (!remoteSessionFinished.get()) {
					remoteSessionFinished.wait();
				}
			}
		} catch (InterruptedException e) {
			throw new IOException(e.toString());
		}
	}

	private class ENDHandler implements MailboxRequestHandler {
		@Override
		public void handleRequest(MailboxRequest request) {
			synchronized (remoteSessionFinished) {
				remoteSessionFinished.set(true);
				remoteSessionFinished.notifyAll();
			}
		}

		@Override
		public MailboxMessage.TYPE getType() {
			return END;
		}

		@Override
		public void protocolFinished() {
		}
	}

	/**
	 * Enables support for SYNC requests between mailbox and client
	 * this runs an incoming session which is fed data from incoming
	 * requests and runs and outgoing Session which writes SYNC requests
	 * <p>
	 * NOTE: This function terminates until the connection is closed
	 */
	protected void enableSYNCHandling() {
		syncHandler = new SYNCHandler();
		mailboxProtocol.registerRequestHandler(syncHandler);
	}

	protected void awaitSYNCHandlerFinished() throws InterruptedException {
		if (syncHandler == null)
			throw new RuntimeException("SYNC handling is not enabled");

		synchronized (syncHandler.isFinished) {
			while (!syncHandler.isFinished.get())
				syncHandler.isFinished.wait();
		}
	}


	private class SYNCHandler implements MailboxRequestHandler {
		private FeedableSyncInputStream syncInputStream;
		private MailboxSyncRequestWriter syncWriter;
		private SyncSession outgoingSession;
		private AtomicBoolean isFinished = new AtomicBoolean(false);

		/*
		 * Note: The stream is not encrypted, since the MailboxProtocol is responsible
		 * for encrypting/decrypting requests
		 */
		private SYNCHandler() {
			syncWriter = new MailboxSyncRequestWriter(mailboxProtocol);
			syncInputStream = new FeedableSyncInputStream();

			// Run an incoming session to handle received SYNC requests
			ioExecutor.execute(() -> {
				try {
					syncSessionFactory
							.createIncomingSession(contactId,
									syncInputStream)
							.run();
				} catch (IOException e) {
					if (LOG.isLoggable(INFO))
						LOG.info(e.toString());
				}
			});
			outgoingSession = syncSessionFactory
					.createDuplexOutgoingSession(contactId,
							transportMaxLatency,
							transportMaxIdleTime, syncWriter);
			// Run an outgoing session to send SYNC requests
			ioExecutor.execute(() -> {
				try {
					outgoingSession.run();
				} catch (IOException e) {
					if (LOG.isLoggable(INFO))
						LOG.info(e.toString());
				}
			});
		}

		/**
		 * Handles an incoming SYNC request. The request contains an unencrypted
		 * BSP stream which is passed to a BSP IncomingSession
		 *
		 * @param request MailboxRequest (SYNC)
		 */
		@Override
		public void handleRequest(MailboxRequest request) {
			MailboxRequestSync syncReq = (MailboxRequestSync) request;

			if (syncReq.isEndOfStream()) {
				syncInputStream.close();
				return;
			}

			try {
				syncInputStream.feed(syncReq.getSyncStream());
			} catch (IOException e) {
				if (LOG.isLoggable(WARNING))
					LOG.warning(e.toString());
			}
		}

		@Override
		public MailboxMessage.TYPE getType() {
			return SYNC;
		}

		@Override
		public void protocolFinished() {
			syncInputStream.close();
			outgoingSession.interrupt();
			LOG.info("SYNC request handler terminated");
			synchronized (isFinished) {
				isFinished.set(true);
				isFinished.notifyAll();
			}
		}
	}

	/**
	 * Get an encrypted Sync stream which can be stored on the mailbox by
	 * issuing a STORE request
	 *
	 * @param c The encrypted sync stream is generated for the given
	 * contactId
	 * @return Encrypted Sync stream or Null if no data is available to send
	 * @throws DbException
	 * @throws IOException
	 */
	@Nullable
	protected byte[] getSyncStreamToStore(ContactId c)
			throws DbException, IOException {
		Transaction txn = null;
		boolean hasMessages;

		try {
			txn = db.startTransaction(true);
			hasMessages = db.hasMessagesOrAcksToSend(txn, c);
			db.commitTransaction(txn);
		} finally {
			if (txn != null)
				db.endTransaction(txn);
		}

		if (!hasMessages)
			return null;

		StreamContext ctx =
				keyManager
						.getStreamContext(c, MailboxConstants.ID);

		if (ctx == null)
			throw new IOException("Could not allocated stream context");

		ByteArrayOutputStream os = new ByteArrayOutputStream();

		StreamWriter streamWriter = streamWriterFactory
				.createStreamWriter(os, ctx);

		syncSessionFactory.createSimplexOutgoingSession(c,
				MailboxConstants.MAX_LATENCY, streamWriter).run();
		byte[] buf = os.toByteArray();

		if (buf.length <= 0)
			return null;

		return buf;
	}

	/**
	 * Reads a Tag from the given InputStream and returns the matching StreamContext
	 * if available
	 *
	 * @param in InputStream to read the tag from
	 * @return StreamContext matching the tag read from the InputStream
	 * @throws IOException
	 * @throws DbException
	 */
	@Nullable
	protected StreamContext readTag(InputStream in)
			throws IOException, DbException {

		byte[] tag = new byte[TAG_LENGTH];
		// read tag from input stream
		int offset = 0;
		while (offset < tag.length) {
			int read = in.read(tag, offset, tag.length - offset);
			if (read == -1) throw new EOFException();
			offset += read;
		}

		StreamContext ctx =
				keyManager.getStreamContext(MailboxConstants.ID, tag);

		if (ctx == null)
			throw new IOException(
					"Received stream with unrecognisable tag");
		return ctx;
	}
}
