package org.briarproject.bramble.mailbox.sessions;

import org.briarproject.bramble.api.contact.ContactId;
import org.briarproject.bramble.api.db.DatabaseComponent;
import org.briarproject.bramble.api.db.DbException;
import org.briarproject.bramble.api.sync.SyncSessionFactory;
import org.briarproject.bramble.api.transport.KeyManager;
import org.briarproject.bramble.api.transport.StreamContext;
import org.briarproject.bramble.api.transport.StreamReaderFactory;
import org.briarproject.bramble.api.transport.StreamWriterFactory;
import org.briarproject.bramble.mailbox.protocol.MailboxMessage;
import org.briarproject.bramble.mailbox.protocol.MailboxProtocol;
import org.briarproject.bramble.mailbox.protocol.MailboxRequest;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestHandler;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestStore;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestTake;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.ProtocolException;
import java.util.concurrent.Executor;
import java.util.logging.Logger;

import static java.util.logging.Level.INFO;
import static java.util.logging.Level.WARNING;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.TAKE;
import static org.briarproject.bramble.util.LogUtils.logException;

/**
 * This session implements a contact communicating with another contact's
 * mailbox
 */
public class ContactMailboxSession extends AbstractMailboxSession {
	private static final Logger LOG =
			Logger.getLogger(ContactMailboxSession.class.getName());
	private final SyncSessionFactory syncSessionFactory;
	private final StreamReaderFactory streamReaderFactory;
	private final MailboxProtocol mailboxProtocol;
	private final ContactId contactId;

	public ContactMailboxSession(ContactId contactId, Executor ioExecutor,
			KeyManager keyManager,
			SyncSessionFactory syncSessionFactory,
			StreamWriterFactory streamWriterFactory,
			StreamReaderFactory streamReaderFactory,
			MailboxProtocol mailboxProtocol,
			int transportMaxLatency, int transportMaxIdleTime,
			DatabaseComponent db) {
		super(ioExecutor, db, keyManager, syncSessionFactory,
				streamWriterFactory,
				streamReaderFactory, mailboxProtocol, transportMaxLatency,
				transportMaxIdleTime, contactId);
		this.syncSessionFactory = syncSessionFactory;
		this.streamReaderFactory = streamReaderFactory;
		this.mailboxProtocol = mailboxProtocol;
		this.contactId = contactId;

		mailboxProtocol.registerRequestHandler(new TAKEHandler());
	}

	@Override
	public void run() {
		// Get messages to send and formulate a STORE request
		byte[] encryptedStream;
		try {
			// Get sync stream if available
			 encryptedStream = getSyncStreamToStore(contactId);
		} catch (IOException | DbException e) {
			logException(LOG, WARNING, e);
			return;
		}

		if (encryptedStream == null)
				return;

		// Send a STORE request and wait for a response
		MailboxRequestStore req = new MailboxRequestStore(encryptedStream);
		try {
			mailboxProtocol.writeRequest(req);
			if (!req.awaitAndGetResponse())
				throw new ProtocolException(req.getError());
		} catch (InterruptedException  | IOException e) {
			// TODO: Handle STORE request error!
			if (LOG.isLoggable(WARNING))
				LOG.info(e.toString());
			return;
		}

		// End the session by issuing and END request and wait for peer
		// to send END request
		try {
			endSession();
		} catch (IOException e) {
			if (LOG.isLoggable(INFO))
				LOG.info(e.toString());
		}
	}


	/**
	 * Handles an incoming TAKE request. The request contains an encrypted BSP
	 * stream. From this stream, the tag is read and a StreamReader is created.
	 * A BSP IncomingSession is created for the stream
	 */
	private class TAKEHandler implements MailboxRequestHandler {

		@Override
		public void handleRequest(MailboxRequest request)
				throws ProtocolException {
			MailboxRequestTake takeRequest = (MailboxRequestTake) request;

			if (takeRequest.hasContactId())
				throw new ProtocolException(
						"TAKE request from contact mailbox must not have contactId");

			InputStream in = new ByteArrayInputStream(
					takeRequest.getEncryptedSyncStream());
			try {
				StreamContext ctx = readTag(in);

				if (!ctx.getContactId().equals(contactId))
					throw new ProtocolException(
							"Contact Id from stream does not match expected contactId!");

				InputStream reader =
						streamReaderFactory.createStreamReader(in, ctx);

				syncSessionFactory
						.createIncomingSession(ctx.getContactId(), reader)
						.run();

			} catch (DbException | IOException e) {
				throw new ProtocolException("Received invalid Stream");
			}

		}

		@Override
		public MailboxMessage.TYPE getType() {
			return TAKE;
		}

		@Override
		public void protocolFinished() {
		}
	}

}
