package org.briarproject.bramble.mailbox.sessions;

import org.briarproject.bramble.api.contact.ContactId;
import org.briarproject.bramble.api.db.DatabaseComponent;
import org.briarproject.bramble.api.mailbox.MailboxStorage;
import org.briarproject.bramble.api.mailbox.MailboxStorageStream;
import org.briarproject.bramble.api.sync.SyncSessionFactory;
import org.briarproject.bramble.api.transport.KeyManager;
import org.briarproject.bramble.api.transport.StreamReaderFactory;
import org.briarproject.bramble.api.transport.StreamWriterFactory;
import org.briarproject.bramble.mailbox.protocol.MailboxMessage;
import org.briarproject.bramble.mailbox.protocol.MailboxProtocol;
import org.briarproject.bramble.mailbox.protocol.MailboxRequest;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestHandler;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestStore;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestTake;

import java.io.IOException;
import java.net.ProtocolException;
import java.util.concurrent.Executor;
import java.util.logging.Logger;

import static java.util.logging.Level.INFO;
import static java.util.logging.Level.WARNING;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.STORE;
import static org.briarproject.bramble.util.LogUtils.logException;

/**
 * This session implements a mailbox communicating with a contact
 * of it's owner (which was previously introduced to the mailbox)
 */
public class MailboxContactSession extends AbstractMailboxSession {
	private static final Logger LOG =
			Logger.getLogger(MailboxContactSession.class.getName());

	private final MailboxProtocol mailboxProtocol;
	private final ContactId contactId;
	private final MailboxStorage mailboxStorage;

	public MailboxContactSession(ContactId contactId, Executor ioExecutor,
			KeyManager keyManager, SyncSessionFactory syncSessionFactory,
			StreamWriterFactory streamWriterFactory,
			StreamReaderFactory streamReaderFactory,
			MailboxProtocol mailboxProtocol, int transportMaxLatency,
			int transportMaxIdleTime, MailboxStorage mailboxStorage,
			DatabaseComponent db) {
		super(ioExecutor, db, keyManager, syncSessionFactory,
				streamWriterFactory,
				streamReaderFactory, mailboxProtocol, transportMaxLatency,
				transportMaxIdleTime, contactId);
		this.mailboxProtocol = mailboxProtocol;
		this.contactId = contactId;
		this.mailboxStorage = mailboxStorage;

		mailboxProtocol.registerRequestHandler(new STOREHandler());
	}

	@Override
	public void run() {
		// Send stored sync streams to connected contact
		try {
			sendStoredStreams();
		} catch (IOException e) {
			if (LOG.isLoggable(WARNING))
				LOG.warning(e.toString());
		} catch (InterruptedException e) {
			logException(LOG, WARNING, e);
			return;
		}

		// End the session by issuing and END request and wait for peer
		// to send END request
		try {
			endSession();
		} catch (IOException e) {
			if (LOG.isLoggable(INFO))
				LOG.info(e.toString());
		}
	}

	void sendStoredStreams() throws IOException, InterruptedException {
		// Get stored sync stream if available
		MailboxStorageStream nextStream =
				mailboxStorage.getStream(contactId, false);

		while (nextStream != null) {
			// Send TAKE request and delete stream if request was successfull
			MailboxRequestTake req =
					new MailboxRequestTake(nextStream.getStream());
			mailboxProtocol.writeRequest(req);

			if (req.awaitAndGetResponse())
				// delete stream if request was successful
				mailboxStorage.deleteStream(nextStream);
			else
				//TODO: stop sending after failed request?
				throw new ProtocolException(req.getError());

			// Get next stored sync stream if available
			nextStream = mailboxStorage.getStream(contactId, false);

		}
	}


	private class STOREHandler implements MailboxRequestHandler {
		@Override
		public void handleRequest(MailboxRequest request)
				throws ProtocolException {
			MailboxRequestStore storeReq = (MailboxRequestStore) request;

			if (storeReq.hasContactId())
				throw new ProtocolException(
						"Contact Id is implicit for contact to mailbox STORE requests");

			mailboxStorage
					.storeStream(contactId, storeReq.getEncryptedSyncStream(),
							true);
		}

		@Override
		public MailboxMessage.TYPE getType() {
			return STORE;
		}

		@Override
		public void protocolFinished() {
		}
	}
}
