package org.briarproject.bramble.mailbox.sessions;

import org.briarproject.bramble.api.contact.ContactId;
import org.briarproject.bramble.api.db.DatabaseComponent;
import org.briarproject.bramble.api.mailbox.MailboxStorage;
import org.briarproject.bramble.api.mailbox.MailboxStorageStream;
import org.briarproject.bramble.api.sync.SyncSessionFactory;
import org.briarproject.bramble.api.transport.KeyManager;
import org.briarproject.bramble.api.transport.StreamReaderFactory;
import org.briarproject.bramble.api.transport.StreamWriterFactory;
import org.briarproject.bramble.mailbox.protocol.MailboxMessage;
import org.briarproject.bramble.mailbox.protocol.MailboxProtocol;
import org.briarproject.bramble.mailbox.protocol.MailboxRequest;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestHandler;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestStore;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestTake;

import java.io.IOException;
import java.net.ProtocolException;
import java.util.concurrent.Executor;
import java.util.logging.Logger;

import static java.util.logging.Level.INFO;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.STORE;

/**
 * This session represents the mailbox side of a connection from the owner
 * of the  mailbox
 */
public class MailboxOwnerSession extends AbstractMailboxSession {

	private static final Logger LOG =
			Logger.getLogger(MailboxOwnerSession.class.getName());
	private final MailboxStorage mailboxStorage;
	private MailboxProtocol mailboxProtocol;

	public MailboxOwnerSession(ContactId contactId, Executor ioExecutor,
			KeyManager keyManager,
			SyncSessionFactory syncSessionFactory,
			StreamWriterFactory streamWriterFactory,
			StreamReaderFactory streamReaderFactory,
			MailboxProtocol mailboxProtocol, int transportMaxLatency,
			int transportMaxIdleTime, MailboxStorage mailboxStorage,
			DatabaseComponent db) {

		super(ioExecutor, db, keyManager, syncSessionFactory,
				streamWriterFactory,
				streamReaderFactory, mailboxProtocol, transportMaxLatency,
				transportMaxIdleTime, contactId);
		this.mailboxProtocol = mailboxProtocol;
		this.mailboxStorage = mailboxStorage;

		enableSYNCHandling();
		mailboxProtocol.registerRequestHandler(new STOREHandler());
	}

	@Override
	public void run() {
		try {
			sendStoredStreams();
			awaitSYNCHandlerFinished();
			// NOTE: SYNC handler only terminates if the connection is closed
			// so it  is not required to end the session manually by
			// sending an END request
		} catch (InterruptedException | IOException e) {
			if (LOG.isLoggable(INFO))
				LOG.info(e.toString());
		}
	}

	void sendStoredStreams() throws IOException, InterruptedException {
		// Get stored sync stream if available
		MailboxStorageStream nextStream =
				mailboxStorage.getStreamForOwner();

		// FIXME: If there are no more streams to send at this time, we should
		// wait
		while (nextStream != null) {
			// Send TAKE request and delete stream if request was successfull
			MailboxRequestTake req =
					new MailboxRequestTake(nextStream.getContactId(), nextStream.getStream());
			mailboxProtocol.writeRequest(req);

			if (req.awaitAndGetResponse())
				// delete stream if request was successful
				mailboxStorage.deleteStream(nextStream);
			else
				//TODO: stop sending after failed request?
				throw new ProtocolException(req.getError());

			// Get next stored sync stream if available
			nextStream = mailboxStorage.getStreamForOwner();
		}
	}

	private class STOREHandler implements MailboxRequestHandler {
		@Override
		public void handleRequest(MailboxRequest request)
				throws ProtocolException {
			MailboxRequestStore storeReq = (MailboxRequestStore) request;

			if (!storeReq.hasContactId())
				throw new ProtocolException(
						"Contact Id must be set for owner to mailbox STORE requests");

			mailboxStorage
					.storeStream(storeReq.getContactId(),
							storeReq.getEncryptedSyncStream(), false);
		}

		@Override
		public MailboxMessage.TYPE getType() {
			return STORE;
		}

		@Override
		public void protocolFinished() {
		}
	}
}
