package org.briarproject.bramble.mailbox.sessions;

import org.briarproject.bramble.api.contact.ContactId;
import org.briarproject.bramble.api.db.DatabaseComponent;
import org.briarproject.bramble.api.db.DbException;
import org.briarproject.bramble.api.mailbox.MailboxInfo;
import org.briarproject.bramble.api.sync.SyncSessionFactory;
import org.briarproject.bramble.api.transport.KeyManager;
import org.briarproject.bramble.api.transport.StreamContext;
import org.briarproject.bramble.api.transport.StreamReaderFactory;
import org.briarproject.bramble.api.transport.StreamWriterFactory;
import org.briarproject.bramble.mailbox.protocol.MailboxMessage;
import org.briarproject.bramble.mailbox.protocol.MailboxProtocol;
import org.briarproject.bramble.mailbox.protocol.MailboxRequest;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestHandler;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestStore;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestTake;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.ProtocolException;
import java.util.concurrent.Executor;
import java.util.logging.Logger;

import static java.util.logging.Level.INFO;
import static java.util.logging.Level.WARNING;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.TAKE;
import static org.briarproject.bramble.util.LogUtils.logException;

/**
 * This session represents an Owner (of a private mailbox) connecting
 * to their mailbox
 */
public class PrivateMailboxSession extends AbstractMailboxSession {
	private static final Logger LOG =
			Logger.getLogger(PrivateMailboxSession.class.getName());
	private final SyncSessionFactory syncSessionFactory;
	private final StreamReaderFactory streamReaderFactory;
	private MailboxProtocol mailboxProtocol;

	public PrivateMailboxSession(ContactId contactId, Executor ioExecutor,
			KeyManager keyManager,
			SyncSessionFactory syncSessionFactory,
			StreamWriterFactory streamWriterFactory,
			StreamReaderFactory streamReaderFactory,
			MailboxProtocol mailboxProtocol, int transportMaxLatency,
			int transportMaxIdleTime, DatabaseComponent db) {

		super(ioExecutor, db, keyManager, syncSessionFactory,
				streamWriterFactory,
				streamReaderFactory, mailboxProtocol, transportMaxLatency,
				transportMaxIdleTime, contactId);
		this.syncSessionFactory = syncSessionFactory;
		this.streamReaderFactory = streamReaderFactory;
		this.mailboxProtocol = mailboxProtocol;

		mailboxProtocol.registerRequestHandler(new TAKEHandler());
		// Register SYNC handler and run outgoing SYNC session
		enableSYNCHandling();
	}

	/**
	 * Stores messages on the private mailbox for a contact
	 * who does not own a mailbox
	 */
	public void handleContactWithoutPrivateMailbox(MailboxInfo mailboxInfo) {
		byte[] encryptedStream;
		try {
			encryptedStream =
					getSyncStreamToStore(mailboxInfo.getContactId());
		} catch (DbException | IOException e) {
			logException(LOG, WARNING, e);
			return;
		}

		if (encryptedStream == null)
			return;
		MailboxRequest req =
				new MailboxRequestStore(mailboxInfo.getAliasId(),
						encryptedStream);
		try {
			mailboxProtocol.writeRequest(req);
			if (!req.awaitAndGetResponse())
				throw new ProtocolException(req.getError());
		} catch (InterruptedException | IOException e) {
			// TODO: Handle STORE error!
			if (LOG.isLoggable(INFO))
				LOG.info(e.toString());
		}
	}

	@Override
	public void run() {
		try {
			awaitSYNCHandlerFinished();
			// NOTE: SYNC handler only terminates if the connection is closed
			// so it  is not required to end the session manually by
			// sending an END request
		} catch (InterruptedException e) {
			logException(LOG, WARNING, e);
		}
	}

	/**
	 * Handles an incoming TAKE request. The request contains an encrypted BSP
	 * stream. From this stream, the tag is read and a StreamReader is created.
	 * A BSP IncomingSession is created for the stream
	 */
	private class TAKEHandler implements MailboxRequestHandler {
		@Override
		public void handleRequest(MailboxRequest request)
				throws ProtocolException {
			MailboxRequestTake takeRequest = (MailboxRequestTake) request;
			InputStream in = new ByteArrayInputStream(
					takeRequest.getEncryptedSyncStream());
			try {
				StreamContext ctx = readTag(in);

				if (ctx.getContactId() != (takeRequest.getContactId()))
					throw new ProtocolException(
							"Stream does not match expected contactId");

				InputStream reader =
						streamReaderFactory.createStreamReader(in, ctx);

				syncSessionFactory
						.createIncomingSession(ctx.getContactId(), reader)
						.run();
			} catch (DbException | IOException e) {
				logException(LOG, WARNING, e);
				throw new ProtocolException("Received invalid stream");
			}
		}

		@Override
		public MailboxMessage.TYPE getType() {
			return TAKE;
		}

		@Override
		public void protocolFinished() {
		}
	}
}
