package org.briarproject.bramble.mailbox.protocol;

import org.briarproject.bramble.api.contact.ContactId;
import org.briarproject.bramble.api.data.BdfReader;
import org.briarproject.bramble.api.data.BdfReaderFactory;
import org.briarproject.bramble.api.data.BdfWriter;
import org.briarproject.bramble.api.data.BdfWriterFactory;
import org.briarproject.bramble.api.lifecycle.IoExecutor;
import org.briarproject.bramble.test.BrambleTestCase;
import org.junit.Before;
import org.junit.Test;

import java.io.IOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.net.ProtocolException;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Executor;
import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.inject.Inject;

import static java.util.concurrent.TimeUnit.SECONDS;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.END;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.STORE;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.SYNC;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.TAKE;
import static org.junit.Assert.assertEquals;

public class MailboxProtocolIntegrationTest extends BrambleTestCase {

	@IoExecutor
	Executor ioExecutor;

	@Inject
	BdfReaderFactory bdfReaderFactory;

	@Inject
	BdfWriterFactory bdfWriterFactory;


	private BdfWriter bdfWriter;
	private BdfReader bdfReader;
	private PipedOutputStream pipedOS;
	private PipedInputStream pipedIS;

	private MailboxProtocol mailboxProtocol;


	public MailboxProtocolIntegrationTest() {
		// The thread pool is unbounded, so use direct handoff
		BlockingQueue<Runnable> queue = new SynchronousQueue<>();
		// Discard tasks that are submitted during shutdown
		RejectedExecutionHandler policy =
				new ThreadPoolExecutor.DiscardPolicy();
		// Create threads as required and keep them in the pool for 60 seconds
		ioExecutor = new ThreadPoolExecutor(0, Integer.MAX_VALUE,
				60, SECONDS, queue, policy);

		MailboxProtocolIntegrationTestComponent component =
				DaggerMailboxProtocolIntegrationTestComponent.builder().build();
		component.inject(this);

	}

	@Before
	public void initialize() throws IOException {

		pipedIS = new PipedInputStream();
		pipedOS = new PipedOutputStream(pipedIS);


		bdfWriter = bdfWriterFactory.createWriter(pipedOS);
		bdfReader = bdfReaderFactory.createReader(pipedIS);
		mailboxProtocol = new MailboxProtocol(ioExecutor, bdfWriter, bdfReader);
	}

	@Test
	public void testProtocolTerminationOnClosedConnection()
			throws IOException, InterruptedException {
		AtomicBoolean finished = new AtomicBoolean(false);

		mailboxProtocol.registerRequestHandler(new MailboxRequestHandler() {
			@Override
			public void handleRequest(MailboxRequest request) {
				synchronized (finished) {
					while (finished.get() != true) {
						try {
							finished.wait();
						} catch (InterruptedException e) {
						}
					}
				}
			}

			@Override
			public MailboxMessage.TYPE getType() {
				return STORE;
			}

			@Override
			public void protocolFinished() {

			}
		});

		ioExecutor.execute(mailboxProtocol);

		MailboxRequest req =
				new MailboxRequestStore(new ContactId(123), "test".getBytes());
		mailboxProtocol.writeRequest(req);

		// close "connection"
		pipedOS.close();

		// ensure that no response is written before IS is closed
		synchronized (finished) {
			finished.set(true);
		}

		// pending request should be marked as failed
		assertEquals(false, req.awaitAndGetResponse());
		assertEquals("Connection closed", req.getError());
	}


	@Test
	public void errorResponse() throws InterruptedException, IOException {
		mailboxProtocol.registerRequestHandler(new MailboxRequestHandler() {
			@Override
			public void handleRequest(MailboxRequest request)
					throws ProtocolException {
				throw new ProtocolException("Error");
			}

			@Override
			public MailboxMessage.TYPE getType() {
				return TAKE;
			}

			@Override
			public void protocolFinished() {
			}
		});


		ioExecutor.execute(mailboxProtocol);
		// Test Error Response
		MailboxRequestTake req =
				new MailboxRequestTake("Test".getBytes());
		mailboxProtocol.writeRequest(req);

		assertEquals(false, req.awaitAndGetResponse());
		assertEquals("java.net.ProtocolException: Error", req.getError());
		pipedOS.close();
	}

	@Test
	public void takeRequest() throws IOException, InterruptedException {
		// Generate and write TAKE request
		MailboxRequestTake req =
				new MailboxRequestTake("Test".getBytes());

		mailboxProtocol.registerRequestHandler(new MailboxRequestHandler() {
			@Override
			public void handleRequest(MailboxRequest request) {
				assertEquals(req.getId(), request.getId());
				assertEquals(TAKE, request.getType());
				assertEquals("Test", new String(req.getEncryptedSyncStream()));
			}

			@Override
			public MailboxMessage.TYPE getType() {
				return TAKE;
			}

			@Override
			public void protocolFinished() {
			}
		});

		ioExecutor.execute(mailboxProtocol);

		mailboxProtocol.writeRequest(req);

		assertEquals(true, req.awaitAndGetResponse());
		pipedOS.close();
	}

	@Test
	public void endRequest() throws IOException, InterruptedException {
		MailboxRequest req = new MailboxRequestEnd();
		mailboxProtocol.registerRequestHandler(new MailboxRequestHandler() {
			@Override
			public void handleRequest(MailboxRequest request) {

				assertEquals(req.getId(), request.getId());
				assertEquals(END, request.getType());
			}

			@Override
			public MailboxMessage.TYPE getType() {
				return END;
			}

			@Override
			public void protocolFinished() {
			}
		});


		ioExecutor.execute(mailboxProtocol);

		mailboxProtocol.writeRequest(req);
		pipedOS.close();
	}

	@Test
	public void storeRequestWithId() throws IOException, InterruptedException {
		// Generate and write TAKE request
		MailboxRequestStore req =
				new MailboxRequestStore(new ContactId(123), "Test".getBytes());

		mailboxProtocol.registerRequestHandler(new MailboxRequestHandler() {
			@Override
			public void handleRequest(MailboxRequest request) {
				MailboxRequestStore recvReq = (MailboxRequestStore) request;
				assertEquals(req.getId(), recvReq.getId());
				assertEquals(req.getType(), recvReq.getType());
				assertEquals(true, recvReq.hasContactId());
				assertEquals(req.getContactId(), recvReq.getContactId());
				assertEquals("Test", new String(req.getEncryptedSyncStream()));
			}

			@Override
			public MailboxMessage.TYPE getType() {
				return STORE;
			}

			@Override
			public void protocolFinished() {
			}
		});

		ioExecutor.execute(mailboxProtocol);

		mailboxProtocol.writeRequest(req);
		assertEquals(true, req.awaitAndGetResponse());

		pipedOS.close();
	}

	@Test
	public void storeRequestWithoutId() throws IOException, InterruptedException {
		// Generate and write TAKE request
		MailboxRequestStore req =
				new MailboxRequestStore("Test".getBytes());

		mailboxProtocol.registerRequestHandler(new MailboxRequestHandler() {
			@Override
			public void handleRequest(MailboxRequest request) {
				MailboxRequestStore recvReq = (MailboxRequestStore) request;
				assertEquals(req.getId(), recvReq.getId());
				assertEquals(req.getType(), recvReq.getType());
				assertEquals(false, recvReq.hasContactId());
				assertEquals("Test", new String(req.getEncryptedSyncStream()));
			}

			@Override
			public MailboxMessage.TYPE getType() {
				return STORE;
			}

			@Override
			public void protocolFinished() {
			}
		});

		ioExecutor.execute(mailboxProtocol);

		mailboxProtocol.writeRequest(req);
		assertEquals(true, req.awaitAndGetResponse());

		pipedOS.close();
	}

	@Test
	public void syncRequest() throws IOException, InterruptedException {
		MailboxRequestSync req =
				new MailboxRequestSync("Test".getBytes());

		mailboxProtocol.registerRequestHandler(new MailboxRequestHandler() {
			@Override
			public void handleRequest(MailboxRequest request) {
				MailboxRequestSync recvReq = (MailboxRequestSync) request;
				assertEquals(req.getId(), recvReq.getId());
				assertEquals(SYNC, recvReq.getType());
				assertEquals("Test", new String(recvReq.getSyncStream()));
				assertEquals(false, recvReq.isEndOfStream());

			}

			@Override
			public MailboxMessage.TYPE getType() {
				return SYNC;
			}

			@Override
			public void protocolFinished() {
			}
		});

		ioExecutor.execute(mailboxProtocol);

		mailboxProtocol.writeRequest(req);

		pipedOS.close();
	}


}