package org.briarproject.mailbox.android;

import android.app.Application;
import android.app.Notification;
import android.app.NotificationChannel;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.content.Context;
import android.content.Intent;
import android.support.annotation.UiThread;
import android.support.v4.app.NotificationCompat;

import org.briarproject.bramble.api.lifecycle.Service;
import org.briarproject.bramble.api.lifecycle.ServiceException;
import org.briarproject.bramble.api.nullsafety.MethodsNotNullByDefault;
import org.briarproject.bramble.api.nullsafety.ParametersNotNullByDefault;
import org.briarproject.mailbox.R;
import org.briarproject.mailbox.android.login.SignInReminderReceiver;
import org.briarproject.mailbox.android.navdrawer.NavDrawerActivity;
import org.briarproject.mailbox.android.splash.SplashScreenActivity;
import org.briarproject.mailbox.api.android.AndroidNotificationManager;

import java.util.concurrent.atomic.AtomicBoolean;

import javax.annotation.concurrent.ThreadSafe;
import javax.inject.Inject;

import static android.app.NotificationManager.IMPORTANCE_LOW;
import static android.content.Context.NOTIFICATION_SERVICE;
import static android.content.Intent.FLAG_ACTIVITY_CLEAR_TOP;
import static android.content.Intent.FLAG_ACTIVITY_NEW_TASK;
import static android.os.Build.VERSION.SDK_INT;
import static android.support.v4.app.NotificationCompat.CATEGORY_SERVICE;
import static android.support.v4.app.NotificationCompat.PRIORITY_LOW;
import static android.support.v4.app.NotificationCompat.PRIORITY_MIN;
import static android.support.v4.app.NotificationCompat.VISIBILITY_SECRET;
import static android.support.v4.content.ContextCompat.getColor;

@ThreadSafe
@MethodsNotNullByDefault
@ParametersNotNullByDefault
class AndroidNotificationManagerImpl implements AndroidNotificationManager,
		Service {

	private final Context appContext;
	private final NotificationManager notificationManager;
	private final AtomicBoolean used = new AtomicBoolean(false);

	// The following must only be accessed on the main UI thread
	private boolean blockSignInReminder = false;

	@Inject
	AndroidNotificationManagerImpl(Application app) {
		appContext = app.getApplicationContext();
		notificationManager = (NotificationManager)
				appContext.getSystemService(NOTIFICATION_SERVICE);
	}

	@Override
	public void startService() throws ServiceException {
		if (used.getAndSet(true)) throw new IllegalStateException();
	}

	@Override
	public void stopService() throws ServiceException {
	}

	@UiThread
	@Override
	public Notification getForegroundNotification() {
		return getForegroundNotification(false);
	}

	@UiThread
	private Notification getForegroundNotification(boolean locked) {
		int title = locked ? R.string.lock_is_locked :
				R.string.ongoing_notification_title;
		int text = locked ? R.string.lock_tap_to_unlock :
				R.string.ongoing_notification_text;
		int icon = locked ? R.drawable.startup_lock :
				R.drawable.notification_ongoing;
		// Ongoing foreground notification that shows BriarService is running
		NotificationCompat.Builder b =
				new NotificationCompat.Builder(appContext,
						ONGOING_CHANNEL_ID);
		b.setSmallIcon(icon);
		b.setColor(getColor(appContext, R.color.briar_primary));
		b.setContentTitle(appContext.getText(title));
		b.setContentText(appContext.getText(text));
		b.setWhen(0); // Don't show the time
		b.setOngoing(true);
		Intent i = new Intent(appContext, NavDrawerActivity.class);
		i.setFlags(FLAG_ACTIVITY_NEW_TASK | FLAG_ACTIVITY_CLEAR_TOP);
		b.setContentIntent(PendingIntent.getActivity(appContext, 0, i, 0));
		if (SDK_INT >= 21) {
			b.setCategory(CATEGORY_SERVICE);
			b.setVisibility(VISIBILITY_SECRET);
		}
		b.setPriority(PRIORITY_MIN);
		return b.build();
	}

	@UiThread
	@Override
	public void updateForegroundNotification(boolean locked) {
		Notification n = getForegroundNotification(locked);
		notificationManager.notify(ONGOING_NOTIFICATION_ID, n);
	}

	@Override
	public void showSignInNotification() {
		if (blockSignInReminder) return;
		if (SDK_INT >= 26) {
			NotificationChannel channel =
					new NotificationChannel(REMINDER_CHANNEL_ID, appContext
							.getString(
									R.string.reminder_notification_channel_title),
							IMPORTANCE_LOW);
			channel.setLockscreenVisibility(
					NotificationCompat.VISIBILITY_SECRET);
			notificationManager.createNotificationChannel(channel);
		}

		NotificationCompat.Builder b =
				new NotificationCompat.Builder(appContext,
						REMINDER_CHANNEL_ID);
		b.setSmallIcon(R.drawable.ic_signout);
		b.setColor(getColor(appContext, R.color.briar_primary));
		b.setContentTitle(
				appContext.getText(R.string.reminder_notification_title));
		b.setContentText(
				appContext.getText(R.string.reminder_notification_text));
		b.setAutoCancel(true);
		b.setWhen(0); // Don't show the time
		b.setPriority(PRIORITY_LOW);

		// Add a 'Dismiss' action
		String actionTitle =
				appContext
						.getString(R.string.reminder_notification_dismiss);
		Intent i1 = new Intent(appContext, SignInReminderReceiver.class);
		i1.setAction(ACTION_DISMISS_REMINDER);
		PendingIntent actionIntent =
				PendingIntent.getBroadcast(appContext, 0, i1, 0);
		b.addAction(0, actionTitle, actionIntent);

		Intent i = new Intent(appContext, SplashScreenActivity.class);
		i.setFlags(FLAG_ACTIVITY_NEW_TASK | FLAG_ACTIVITY_CLEAR_TOP);
		b.setContentIntent(PendingIntent.getActivity(appContext, 0, i, 0));

		notificationManager.notify(REMINDER_NOTIFICATION_ID, b.build());
	}

	@Override
	public void clearSignInNotification() {
		notificationManager.cancel(REMINDER_NOTIFICATION_ID);
	}

	@Override
	public void blockSignInNotification() {
		blockSignInReminder = true;
	}
}
