package org.briarproject.mailbox.android.viewmodels;

import android.app.Application;
import android.arch.lifecycle.MutableLiveData;
import android.support.annotation.NonNull;

import org.briarproject.bramble.api.contact.Contact;
import org.briarproject.bramble.api.contact.ContactId;
import org.briarproject.bramble.api.contact.ContactManager;
import org.briarproject.bramble.api.contact.MailboxOwner;
import org.briarproject.bramble.api.contact.event.ContactAddedEvent;
import org.briarproject.bramble.api.contact.event.ContactRemovedEvent;
import org.briarproject.bramble.api.db.DatabaseExecutor;
import org.briarproject.bramble.api.event.Event;
import org.briarproject.bramble.api.event.EventBus;
import org.briarproject.bramble.api.event.EventListener;
import org.briarproject.bramble.api.lifecycle.LifecycleManager;
import org.briarproject.bramble.api.plugin.ConnectionRegistry;
import org.briarproject.bramble.api.plugin.event.ContactConnectedEvent;
import org.briarproject.bramble.api.plugin.event.ContactDisconnectedEvent;

import java.util.Collection;
import java.util.concurrent.Executor;
import java.util.logging.Logger;

import javax.inject.Inject;

import static java.util.logging.Level.WARNING;
import static org.briarproject.bramble.api.contact.ContactType.MAILBOX_OWNER;
import static org.briarproject.bramble.util.LogUtils.logException;

public class MailboxOwnerStatusViewModel extends BaseViewModel implements
		EventListener {

	public static final String TAG =
			MailboxOwnerStatusViewModel.class.getName();
	private static final Logger LOG = Logger.getLogger(TAG);
	public final MutableLiveData<MailboxOwnerStatus>
			mailboxOwnerStatusLiveData = new MutableLiveData<>();
	@Inject
	volatile ContactManager contactManager;
	@Inject
	volatile ConnectionRegistry connectionRegistry;
	@Inject
	volatile LifecycleManager lifecycleManager;
	@Inject
	@DatabaseExecutor
	Executor databaseExecutor;
	@Inject
	EventBus eventBus;
	private volatile MailboxOwnerStatus mailboxOwnerStatus;

	public MailboxOwnerStatusViewModel(@NonNull Application application) {
		super(application);
		loadOwner();
		eventBus.addListener(this);
	}

	@Override
	protected void onCleared() {
		super.onCleared();
		eventBus.removeListener(this);
	}

	@Override
	void injectViewModel(ViewModelComponent viewModelComponent) {
		viewModelComponent.inject(this);
	}

	@Override
	public void eventOccurred(Event e) {
		if (e instanceof ContactConnectedEvent)
			updateStatus(((ContactConnectedEvent) e).getContactId());
		else if (e instanceof ContactDisconnectedEvent)
			updateStatus(((ContactDisconnectedEvent) e).getContactId());
		else if (e instanceof ContactAddedEvent) {
			if (mailboxOwnerStatus == null)
				loadOwner();
		} else if (e instanceof ContactRemovedEvent) {
			if (mailboxOwnerStatus != null &&
					((ContactRemovedEvent) e).getContactId()
							.equals(mailboxOwnerStatus.getId())) {
				mailboxOwnerStatus = null;
				mailboxOwnerStatusLiveData.postValue(mailboxOwnerStatus);
			}
		}

	}

	private void updateStatus(ContactId contactId) {
		if (mailboxOwnerStatus != null &&
				mailboxOwnerStatus.getId().equals(contactId)) {
			mailboxOwnerStatus
					.setOnline(connectionRegistry.isConnected(contactId));
			mailboxOwnerStatusLiveData.postValue(mailboxOwnerStatus);
		}
	}

	private void loadOwner() {
		databaseExecutor.execute(() -> {
			try {
				lifecycleManager.waitForDatabase();
				Collection<Contact> result =
						contactManager.getContactsByType(MAILBOX_OWNER);
				if (!result.isEmpty()) {
					MailboxOwner owner =
							(MailboxOwner) result.iterator().next();
					mailboxOwnerStatus =
							new MailboxOwnerStatus(owner, getStatus(owner));
					mailboxOwnerStatusLiveData.postValue(mailboxOwnerStatus);
				}
			} catch (Exception e) {
				logException(LOG, WARNING, e);
			}
		});
	}

	private boolean getStatus(Contact owner) {
		return connectionRegistry.isConnected(owner.getId());
	}

	public class MailboxOwnerStatus {
		private final MailboxOwner mailboxOwner;
		private boolean online;

		MailboxOwnerStatus(MailboxOwner mailboxOwner, boolean online) {
			this.mailboxOwner = mailboxOwner;
			this.online = online;
		}

		public String getName() {
			return mailboxOwner.getAuthor().getName();
		}

		public ContactId getId() {
			return mailboxOwner.getId();
		}

		public boolean isOnline() {
			return online;
		}

		public void setOnline(boolean isOnline) {
			online = isOnline;
		}
	}
}
